@@ -39,10 +39,11 @@ __credits__ = [
     'Jennifer Snyder <jsnyder@yelp.com>',
 ]
 
-__version__ = '0.2.1'
+__version__ = '0.3.0'
 
 import decimal
 import inspect
+import optparse
 import sys
 
 #: One hour, in seconds
@@ -54,6 +55,12 @@ ONE_DAY = 60 * 60 * 24
 #: One week, in seconds
 ONE_WEEK = 60 * 60 * 24 * 7
 
+#: Default ID type for doloop tables
+DEFAULT_ID_TYPE = 'int'
+
+#: Default storage engine for doloop tables
+DEFAULT_STORAGE_ENGINE = 'InnoDB'
+
 
 ### MySQL module compabitibility ###
 
@@ -63,49 +70,6 @@ ONE_WEEK = 60 * 60 * 24 * 7
 # Refer to the DBI specification (PEP 249) here:
 # from http://www.python.org/dev/peps/pep-0249/
 
-# MySQL error codes. From
-# http://dev.mysql.com/doc/refman/5.5/en/error-messages-server.html
-_LOCK_WAIT_TIMEOUT = 1205
-_LOCK_DEADLOCK = 1213
-
-_RECOVERABLE_MYSQL_ERROR_CODES = set([
-    _LOCK_WAIT_TIMEOUT,
-    _LOCK_DEADLOCK,
-])
-
-# Pretty much every module does things differently, so all we have to go
-# on are the class names laid out in PEP 249.
-_DBI_EXCEPTION_CLASS_NAMES = set([
-    'InterfaceError',
-    'DatabaseError',
-    'DataError',
-    'OperationalError',
-    'IntegrityError',
-    'InternalError',
-    'ProgrammingError',
-    'NotSupportedError',
-])
-
-
-def _is_db_exception(e):
-    """Check if an exception is plausibly from a DBI database driver, based on
-    its name."""
-    return any(cls.__name__ in _DBI_EXCEPTION_CLASS_NAMES
-               for cls in inspect.getmro(e.__class__))
-
-
-def _is_recoverable(e):
-    """Check if an exception is a recoverable MySQL exception."""
-    if not _is_db_exception(e):
-        return
-
-    if hasattr(e, 'errno'):  # mysql.connector
-        return e.errno in _RECOVERABLE_MYSQL_ERROR_CODES
-    else:
-        return any(arg in _RECOVERABLE_MYSQL_ERROR_CODES
-                   for arg in e.args or ())
-
-
 def _paramstyle(cursor):
     """Figure out the paramstyle (e.g. qmark, format) used by the
     given database cursor. DBI only specifies that paramstyle needs to be
@@ -164,6 +128,9 @@ def _execute(cursor, qmark_query, params):
     not correctly handle question marks in string literals or double question
     marks.
     """
+    # make sure we haven't lapsed into a different paramstyle
+    assert '%s' not in qmark_query
+
     paramstyle = _paramstyle(cursor)
     format_query = qmark_query.replace('?', '%s')
 
@@ -205,56 +172,58 @@ def _to_list(x):
         return [x]
 
 
-def _run(query, dbconn, level='REPEATABLE READ', read_only=False, retry=True):
-    """Do a query in a transaction.
+def _run(query, dbconn, lock=None, level=None, read_only=False):
+    """Run a query, optionally locking a tables. If an exception
+    is thrown, we unlock the tables and roll back the transaction
+    before re-raising the exception.
 
     :param dbconn: any DBI-compliant MySQL connection object
     :param query: a function which takes a db cursor as its only argument
-    :param string level: MySQL transaction isolation level
-    :param bool read_only: rollback after the transaction
-    :param bool retry: retry on deadlock or lock wait timeout
+    :param lock: if set, we'll set ``autocommit`` to 0, and lock that table in
+                 ``WRITE`` mode before running the query. If not set,
+                 we'll run the query in a transaction instead.
+    :param str level: transaction isolation level to use. Mutually exclusive
+                      with *lock*.
+    :param bool read_only: if true, roll back after running the query
+                           (otherwise we commit)
 
-    If there is already a transaction in progress (i.e. you're using
-    *dbconn* for non-:py:mod:`doloop` things), it'll be rolled back.
+    If there is already a transaction in progress on *dbconn*, we'll roll
+    it back, and (if *lock* is set) unlock any tables currently locked. If
+    something goes wrong during the query, we'll roll back and unlock any
+    tables we locked.
     """
-    return _run_with_rollbacks(lambda cursor, _: query(cursor),
-                               dbconn, level=level, read_only=read_only,
-                               retry=retry)
+    dbconn.rollback()
 
+    cursor = dbconn.cursor()
 
-def _run_with_rollbacks(query, dbconn, level='REPEATABLE READ',
-                        read_only=False, retry=True):
-    """Like :py:func:`~doloop._run`, except query takes an additional
-    argument which is a function that can be called to roll back and restart
-    the transaction.
-    """
-    def rollback_and_restart_transaction():
+    try:
+        if lock:
+            assert not level
+            cursor.execute('UNLOCK TABLES')
+
+            cursor.execute('SET autocommit = 0')
+            cursor.execute('LOCK TABLES `%s` WRITE' % lock)
+        else:
+            if level:
+                cursor.execute('SET TRANSACTION ISOLATION LEVEL ' + level)
+            cursor.execute('START TRANSACTION')
+
+        result = query(cursor)
+
+        if read_only:
+            dbconn.rollback()
+        else:
+            dbconn.commit()
+
+        return result
+
+    except:
         dbconn.rollback()
+        raise
 
-        cursor = dbconn.cursor()
-        cursor.execute('SET TRANSACTION ISOLATION LEVEL ' + level)
-        cursor.execute('START TRANSACTION')
-
-    while True:
-        try:
-            try:
-                rollback_and_restart_transaction()
-
-                result = query(
-                    dbconn.cursor(), rollback_and_restart_transaction)
-
-                if read_only:
-                    dbconn.rollback()
-                else:
-                    dbconn.commit()
-
-                return result
-            except:
-                dbconn.rollback()
-                raise
-        except Exception, e:
-            if not (retry and _is_recoverable(e)):
-                raise
+    finally:
+        if lock:
+            cursor.execute('UNLOCK TABLES')
 
 
 def _check_table_is_a_string(table):
@@ -265,7 +234,8 @@ def _check_table_is_a_string(table):
 
 ### Creating a task loop ###
 
-def create(dbconn, table, id_type='INT'):
+def create(dbconn, table, id_type=DEFAULT_ID_TYPE,
+           engine=DEFAULT_STORAGE_ENGINE):
     """Create a task loop table. It has a schema like this:
 
     .. code-block:: sql
@@ -291,16 +261,18 @@ def create(dbconn, table, id_type='INT'):
     :param str table: name of your task loop table. Something ending in
                       ``_loop`` is recommended.
     :param str id_type: alternate type for the ``id`` field (e.g.
-                        ``'VARCHAR(64)``')
+                        ``'VARCHAR(64)'``)
+    :param str engine: alternate storage engine to use (e.g. ``'MyISAM'``)
 
     There is no ``drop()`` function because programmatically dropping tables
     is risky. The relevant SQL is just ``DROP TABLE `...```.
     """
-    sql = sql_for_create(table, id_type=id_type)
+    sql = sql_for_create(table, id_type=id_type, engine=engine)
     dbconn.cursor().execute(sql)
 
 
-def sql_for_create(table, id_type='INT'):
+def sql_for_create(table, id_type=DEFAULT_ID_TYPE,
+                   engine=DEFAULT_STORAGE_ENGINE):
     """Get SQL used by :py:func:`create`.
 
     Useful to power :command:`create-doloop-table` (included with this
@@ -315,7 +287,36 @@ def sql_for_create(table, id_type='INT'):
     `lock_until` INT default NULL,
     PRIMARY KEY (`id`),
     KEY `lock_until` (`lock_until`, `last_updated`)
-) ENGINE=InnoDB""" % (table, id_type)
+) ENGINE=%s""" % (table, id_type, engine)
+
+
+def _main_for_create_doloop_table(args):
+    """Driver for the create-doloop-table script. See docs/scripts.rst
+    for details."""
+    usage = '%prog [options] table [table ...] | mysql -D dbname'
+    description = ('Print SQL to create one or more task loop tables.')
+    parser = optparse.OptionParser(usage=usage, description=description)
+
+    parser.add_option(
+        '-i', '--id-type', dest='id_type',
+        default=DEFAULT_ID_TYPE,
+        help='Type for the ID field (default: %default)')
+
+    parser.add_option(
+        '-e', '--engine', dest='engine',
+        default=DEFAULT_STORAGE_ENGINE,
+        help='Type for the ID field (default: %default)')
+
+    options, tables = parser.parse_args(args)
+
+    if not tables:
+        parser.error('You must specify at least one table name')
+
+    for table in tables:
+        print sql_for_create(table,
+                             id_type=options.id_type,
+                             engine=options.engine) + ';'
+        print
 
 
 ### Adding and removing IDs ###
@@ -333,8 +334,7 @@ def add(dbconn, table, id_or_ids, updated=False, test=False):
 
     :return: number of IDs that are new
 
-    Runs this query in ``REPEATABLE READ`` mode, retrying on deadlock or lock
-    wait timeout:
+    Runs this query with a write lock on *table*:
 
     .. code-block:: sql
 
@@ -352,7 +352,7 @@ def add(dbconn, table, id_or_ids, updated=False, test=False):
     def query(cursor):
         return _add(cursor, table, ids, updated=updated)
 
-    return _run(query, dbconn, read_only=test)
+    return _run(query, dbconn, lock=table, read_only=test)
 
 
 def _add(cursor, table, ids, updated=False):
@@ -392,8 +392,7 @@ def remove(dbconn, table, id_or_ids, test=False):
 
     :return: number of IDs removed
 
-    Runs this query in ``REPEATABLE READ`` mode, retrying on deadlock or lock
-    wait timeout:
+    Runs this query with a write lock on *table*:
 
     .. code-block:: sql
 
@@ -409,10 +408,11 @@ def remove(dbconn, table, id_or_ids, test=False):
         table, ', '.join('?' for _ in ids))
 
     def query(cursor):
+        cursor.execute('LOCK TABLE `%s` WRITE' % table)
         _execute(cursor, sql, ids)
         return cursor.rowcount
 
-    return _run(query, dbconn, read_only=test)
+    return _run(query, dbconn, lock=table, read_only=test)
 
 
 ### Getting and updating IDs ###
@@ -451,8 +451,7 @@ def get(dbconn, table, limit, lock_for=ONE_HOUR, min_loop_time=ONE_HOUR,
 
     :return: list of IDs
 
-    Runs this query in ``REPEATABLE READ`` mode, retrying on deadlock or lock
-    wait timeout:
+    Runs this query with a write lock on *table*:
 
     .. code-block:: sql
 
@@ -460,7 +459,6 @@ def get(dbconn, table, limit, lock_for=ONE_HOUR, min_loop_time=ONE_HOUR,
             WHERE `lock_until` <= UNIX_TIMESTAMP()
             ORDER BY `lock_until`, `last_updated`
             LIMIT ...
-            FOR UPDATE
 
         SELECT `id` FROM `...`
             WHERE `lock_until` IS NULL
@@ -468,14 +466,9 @@ def get(dbconn, table, limit, lock_for=ONE_HOUR, min_loop_time=ONE_HOUR,
                  OR `last_updated` <= UNIX_TIMESTAMP() - ...)
             ORDER BY `last_updated`
             LIMIT ...
-            FOR UPDATE
 
         UPDATE `...` SET `lock_until` = UNIX_TIMESTAMP() + ...
             WHERE `id` IN (...)
-
-    If the first query returns no rows, we roll back before continuing to
-    avoid holding on to a gap lock (we only need to lock actual rows).
-    Otherwise, all three queries are run in a single transaction.
     """
     # do type-checking up front, to avoid cryptic MySQL errors
 
@@ -507,16 +500,14 @@ def get(dbconn, table, limit, lock_for=ONE_HOUR, min_loop_time=ONE_HOUR,
     select_bumped = ('SELECT `id` FROM `%s`'
                      ' WHERE `lock_until` <= UNIX_TIMESTAMP()'
                      ' ORDER BY `lock_until`, `last_updated`'
-                     ' LIMIT ?'
-                     ' FOR UPDATE') % (table,)
+                     ' LIMIT ?') % (table,)
 
     select_unlocked = ('SELECT `id` FROM `%s`'
                        ' WHERE `lock_until` IS NULL'
                        ' AND (`last_updated` IS NULL OR'
                        ' `last_updated` <= UNIX_TIMESTAMP() - ?)'
                        ' ORDER BY `last_updated`'
-                       ' LIMIT ?'
-                       ' FOR UPDATE') % (table,)
+                       ' LIMIT ?') % (table,)
 
     # this is a function because we need to know how many IDs there are
     def update_sql(ids):
@@ -524,17 +515,12 @@ def get(dbconn, table, limit, lock_for=ONE_HOUR, min_loop_time=ONE_HOUR,
                 ' WHERE `id` IN (%s)' %
                 (table, ', '.join('?' for _ in ids)))
 
-    def query(cursor, rollback_and_restart_transaction):
+    def query(cursor):
         ids = []
 
         _execute(cursor, select_bumped, [limit])
         ids.extend(row[0] for row in cursor.fetchall())
 
-        # if there are no locked IDs (the common case), don't hold on to
-        # the gap lock; we only need to lock actual rows
-        if not ids:
-            rollback_and_restart_transaction()
-
         if len(ids) < limit:
             _execute(cursor, select_unlocked,
                      [min_loop_time, limit - len(ids)])
@@ -547,7 +533,7 @@ def get(dbconn, table, limit, lock_for=ONE_HOUR, min_loop_time=ONE_HOUR,
 
         return ids
 
-    return _run_with_rollbacks(query, dbconn, read_only=test)
+    return _run(query, dbconn, lock=table, read_only=test)
 
 
 def did(dbconn, table, id_or_ids, auto_add=True, test=False):
@@ -565,8 +551,7 @@ def did(dbconn, table, id_or_ids, auto_add=True, test=False):
 
     :return: number of rows updated (mostly useful as a sanity check)
 
-    Runs this query in ``REPEATABLE READ`` mode, retrying on deadlock or lock
-    wait timeout:
+    Runs this query with a write lock on *table*:
 
     .. code-block:: sql
 
@@ -597,7 +582,7 @@ def did(dbconn, table, id_or_ids, auto_add=True, test=False):
         _execute(cursor, sql, ids)
         return cursor.rowcount
 
-    return _run(query, dbconn, read_only=test)
+    return _run(query, dbconn, lock=table, read_only=test)
 
 
 def unlock(dbconn, table, id_or_ids, auto_add=True, test=False):
@@ -618,8 +603,7 @@ def unlock(dbconn, table, id_or_ids, auto_add=True, test=False):
              are configured.) Don't use this for anything more critical than
              sanity checks and logging.
 
-    Runs this query in ``REPEATABLE READ`` mode, retrying on deadlock or lock
-    wait timeout:
+    Runs this query with a write lock on *table*:
 
     .. code-block:: sql
 
@@ -662,7 +646,7 @@ def unlock(dbconn, table, id_or_ids, auto_add=True, test=False):
 
         return rowcount
 
-    return _run(query, dbconn, read_only=test)
+    return _run(query, dbconn, lock=table, read_only=test)
 
 
 ### Prioritization ###
@@ -695,8 +679,7 @@ def bump(dbconn, table, id_or_ids, lock_for=0, auto_add=True, test=False):
 
     :return: number of IDs bumped (mostly useful as a sanity check)
 
-    Runs this query in ``REPEATABLE READ`` mode, retrying on deadlock or lock
-    wait timeout:
+    Runs this query with a write lock on *table*:
 
     .. code-block:: sql
 
@@ -733,7 +716,7 @@ def bump(dbconn, table, id_or_ids, lock_for=0, auto_add=True, test=False):
         _execute(cursor, sql, [lock_for, lock_for] + ids)
         return cursor.rowcount
 
-    return _run(query, dbconn, read_only=test)
+    return _run(query, dbconn, lock=table, read_only=test)
 
 
 ### Auditing ###
@@ -749,10 +732,9 @@ def check(dbconn, table, id_or_ids):
     locked_for)``, that is, the current time minus ``last_updated``, and
     ``lock_for`` minus the current time (both of these in seconds).
 
-    This function does not require write access to your database.
-
-    Runs this query in ``REPEATABLE READ`` mode, retrying on deadlock or lock
-    wait timeout:
+    This function does not require write access to your database. It runs
+    the following query in a transaction (without setting transaction
+    isolation level):
 
     .. code-block:: sql
 
@@ -779,7 +761,7 @@ def check(dbconn, table, id_or_ids):
         return dict((id_, (since_updated, locked_for))
                     for id_, since_updated, locked_for in cursor.fetchall())
 
-    return _run(query, dbconn, level='READ COMMITTED', read_only=True)
+    return _run(query, dbconn, read_only=True)
 
 
 def stats(dbconn, table, delay_thresholds=None):
@@ -922,8 +904,7 @@ def stats(dbconn, table, delay_thresholds=None):
 
         return r
 
-    return _run(query, dbconn, level='READ UNCOMMITTED', read_only=True,
-                retry=False)  # no locking, so no plausible need to retry
+    return _run(query, dbconn, level='READ UNCOMMITTED', read_only=True)
 
 
 ### Object-Oriented version ###
@@ -954,7 +935,7 @@ class DoLoop(object):
                        any caching/pooling/etc. inside your callable.
         :param string table: name of your task loop table
 
-        You can read (but not change) the table name by calling ``self.table``
+        You can read (but not change) the table name from ``self.table``
         """
         if hasattr(dbconn, '__call__'):
             self._make_dbconn = dbconn
